"""Environment information."""

import os
import traceback
from typing import cast

import requests

# The name of the environment variable that determines the environment
_env_key = 'APP_ENV'

_dev_key = 'dev'
_test_key = 'test'
_integration_key = 'integration'
_e2e_key = 'e2e'
_prod_key = 'production'


def env() -> str:
    if _env_key in os.environ:
        return cast(str, os.environ.get(_env_key)).strip()
    return _dev_key


def is_dev() -> bool:
    return env() == _dev_key


def is_test() -> bool:
    return env() == _test_key


def is_integration() -> bool:
    return env() == _integration_key


def is_e2e() -> bool:
    return env() == _e2e_key


def is_prod() -> bool:
    return env() == _prod_key


_is_google_cloud = None


def is_google_cloud() -> bool:
    """GCP machines have access to an internal metadata server - if the server is not available, you're not on GCP.

    https://cloud.google.com/compute/docs/storing-retrieving-metadata#querying

    Returns:
        bool: True if running on GCP.
    """
    global _is_google_cloud

    if _is_google_cloud is not None:
        return _is_google_cloud

    # There are a LOT of linting exceptions in this function, as WPS **really** doesn't like
    # global variables, but the `_is_google_cloud` variable, by definition, contains
    # global immutable state (the module is either running on GCP or it isn't, that's not going
    # to change), we just don't know the state until we check the first time.

    try:
        response = requests.get(
            'http://metadata.google.internal/', timeout=1.5, headers={'Metadata-Flavor': 'Google'},  # noqa: WPS432
        )
        _is_google_cloud = response.status_code == 200  # noqa: WPS432,WPS122,WPS442
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
        _is_google_cloud = False  # noqa: WPS432,WPS121,WPS122,WPS442

    return _is_google_cloud  # noqa: WPS432,WPS121,WPS122,WPS442,R504


# We don't try to test this
def is_ipython() -> bool:  # pragma: no cover
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


# This isn't an exact science, but we can try!
def is_pytest() -> bool:
    # Shortcut that only works when you're actually in a test
    if 'PYTEST_CURRENT_TEST' in os.environ:
        return True

    for frame in traceback.extract_stack():  # pragma: no cover
        # The pytest can appear as `site-packages/_pytest` or `site-packages/pytest`
        if 'site-packages' in frame.filename and 'pytest' in frame.filename:
            return True

    return False  # pragma: no cover



"""Pre-condition check decorator."""

from asyncio import iscoroutinefunction
from typing import Callable

from asgiref.sync import async_to_sync
from makefun import wraps


def pre_condition(pre_condition_fn: Callable[..., bool]):  # noqa: WPS231,WPS212
    """Executes a pre-condition function over the provided arguments before calling the decorated function.

    Automatically handles async pre-condition functions and wrapped functions.

    Exceptions raised in the pre-condition will pass through the decorator.
    If the pre-condition returns False, an UnmetPreconditionException will be raised.

    Args:
        pre_condition_fn (Callable[..., bool]): The pre-condition function.

    Raises:
        UnmetPreconditionException: If the pre-condition fails.

    Returns:
        Callable: The wrapped function
    """

    def pre_condition_decorator(fn: Callable):  # noqa: WPS231
        if iscoroutinefunction(fn) and iscoroutinefunction(pre_condition_fn):

            @wraps(fn)
            async def with_precondition(*args, **kwargs):
                if not await pre_condition_fn(*args, **kwargs):
                    raise UnmetPreconditionException
                return await fn(*args, **kwargs)

        elif iscoroutinefunction(fn):

            @wraps(fn)
            async def with_precondition(*args, **kwargs):  # noqa: WPS440
                if not pre_condition_fn(*args, **kwargs):
                    raise UnmetPreconditionException
                return await fn(*args, **kwargs)

        elif iscoroutinefunction(pre_condition_fn):

            @wraps(fn)
            def with_precondition(*args, **kwargs):  # noqa: WPS440
                # Use async_to_sync to call an async from a sync: https://www.aeracode.org/2018/02/19/python-async-simplified/
                if not async_to_sync(pre_condition_fn)(*args, **kwargs):
                    raise UnmetPreconditionException
                return fn(*args, **kwargs)

        else:

            @wraps(fn)
            def with_precondition(*args, **kwargs):  # noqa: WPS440
                if not pre_condition_fn(*args, **kwargs):
                    raise UnmetPreconditionException
                return fn(*args, **kwargs)

        return with_precondition

    return pre_condition_decorator


class UnmetPreconditionException(Exception):
    ...



"""Feature flag management."""

import re
import warnings
from typing import Dict

from outcome.utils import env
from outcome.utils.config import Config
from rich.console import Console
from rich.table import Table


class FeatureException(Exception):
    ...


_feature_pattern = r'^[a-z]+(_?[a-z]+_?[a-z])*$'
_feature_prefix = 'WITH_FEAT_'

_features: Dict[str, bool]
_config: Config


def reset():
    global _config, _features
    _config = Config()  # noqa: WPS122,WPS442
    _features = {}  # noqa: WPS122,WPS442


reset()


def _is_valid_feature_name(feature: str):
    return all(re.match(_feature_pattern, part) is not None for part in feature.split('.'))


def _feature_to_env_key(feature: str) -> str:
    replaced = feature.replace('.', '_').upper()
    return f'{_feature_prefix}{replaced}'


def _coerce_boolean(v):
    return str(v).lower() in {'yes', 'y', 'true', 't', '1'}


def set_config(config: Config):
    global _config
    _config = config  # noqa: WPS122,WPS442


def register_feature(feature: str, default: bool = False) -> None:
    if not _is_valid_feature_name(feature):
        raise FeatureException(f'Invalid feature name: {feature}')

    if feature in _features:
        raise FeatureException(f'Duplicate feature: {feature}')

    _features[feature] = default


def features() -> Dict[str, bool]:
    return {k: is_active(k) for k in _features.keys()}


def display_features():  # pragma: no cover
    console = Console()

    table = Table(show_header=True, header_style='bold')
    table.add_column('Feature Flag')
    table.add_column('State', justify='right')
    table.add_column('Default State', justify='right')

    def state_repr(state):
        return '[bold green]active[/bold green]' if state else '[bold red]inactive[/bold red]'

    for feat, state in features().items():
        table.add_row(feat, state_repr(state), state_repr(_features[feat]))

    console.print(table)


def is_active(feature: str) -> bool:
    if not _feature_check(feature):
        return False

    try:
        feature_key = _feature_to_env_key(feature)
        value = _config.get(feature_key)
        return _coerce_boolean(value) if isinstance(value, str) else value
    except KeyError:
        ...

    return _features[feature]


def _feature_check(feature):
    if feature not in _features:
        if env.is_dev():
            warnings.warn(
                f'Checking unknown feature "{feature}", maybe you forgot to register it? This will raise an exception in production',  # noqa: E501
                RuntimeWarning,
            )
            return False
        raise FeatureException(f'Unknown feature: {feature}')
    return True


def set_feature_default(feature: str, default_state: bool) -> None:
    _feature_check(feature)
    _features[feature] = default_state



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['outcome', 'outcome.utils', 'outcome.utils.bin', 'outcome.utils.jinja2']

package_data = \
{'': ['*']}

install_requires = \
['asgiref>=3.2.10,<4.0.0',
 'cachetools>=4.1.1,<5.0.0',
 'colored>=1.4.2,<2.0.0',
 'dogpile.cache>=1.0.2,<2.0.0',
 'jinja2>=2.11.2,<3.0.0',
 'makefun>=1.9.3,<2.0.0',
 'pyfakefs>=4.1.0,<5.0.0',
 'requests>=2.24.0,<3.0.0',
 'rich>=6.2,<10.0',
 'semver>=2.10.2,<3.0.0',
 'toml>=0.10.1,<0.11.0']

entry_points = \
{'console_scripts': ['otc-utils = outcome.utils.bin.otc_utils:main']}

setup_kwargs = {
    'name': 'outcome-utils',
    'version': '4.16.0',
    'description': 'A collection of python utils.',
    'long_description': "# utils-py\n![ci-badge](https://github.com/outcome-co/utils-py/workflows/Release/badge.svg?branch=v4.16.0) ![version-badge](https://img.shields.io/badge/version-4.16.0-brightgreen)\n\nA set of python utilities.\n\n## Usage\n\n```sh\npoetry add outcome-utils\n```\n\n### Cache\n\nTo add cache to a module\n``` python\nfrom outcome.utils import cache\n\ncache_settings = {\n    '<your_prefix>.expiration': 300,  # Default\n    '<your_prefix>.backend': 'memory',  # Default\n}\n\nregion = cache.get_cache_region()\ncache.configure_cache_region(region, settings=cache_settings, prefix='<your_prefix>')\n```\n\nThen add to the functions to cache:\n``` python\n@region.cache_on_arguments()\ndef func_to_cache():\n    ...\n```\n\nOr for async functions:\n``` python\n@region.cache_on_arguments()\n@cache.cache_async\nasync def async_func_to_cache():\n    ...\n```\n\nTo have the cache persist on disk, specify the path\n``` python\nfrom pathlib import Path\n\ncache_settings = {\n    ...\n    '<your_prefix>.cache_path': f'{Path.home()}/.cache/example_path/cache.pkl'',\n    ...\n}\n```\n\n## Development\n\nRemember to run `./pre-commit.sh` when you clone the repository.\n",
    'author': 'Douglas Willcocks',
    'author_email': 'douglas@outcome.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/outcome-co/utils-py',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.6,<3.9.0',
}


setup(**setup_kwargs)



"""Helpers to print to the console."""

import sys
import traceback
from contextlib import contextmanager
from typing import IO, Generator, List, Optional, Union

import colored
from colored import stylize

# Weights
_dim = 'dim'
_bold = 'bold'

# Colors
_green = 'green'
_red = 'red'


# Messages
_failure = 'Failure'
_success = 'Success'
_skipped = 'Skipped'
_error = 'Error'
_task_failed = 'Task Failed'


# Default
_bullet = '•'

_raise = 'raise'
_print = 'print'
_exit = 'exit'
_print_and_exit = 'print_and_exit'


def bold_red(message: str) -> str:
    return stylize(message, colored.fg(_red) + colored.attr(_bold))


def bold_green(message: str) -> str:
    return stylize(message, colored.fg(_green) + colored.attr(_bold))


def dim(message: str) -> str:
    return stylize(message, colored.attr(_dim))


def bold(message: str) -> str:
    return stylize(message, colored.attr(_bold))


class Status:
    """A lightweight object to represent the status of a task."""

    def __init__(self) -> None:
        self.completed = False

    def failure(self, message: Optional[str] = None, detail: Optional[str] = None) -> None:
        """Indicate that the task failed.

        Args:
            message (str, optional): The optional error message to display.
            detail (str, optional): Optional detail to appear under the task status.
        """
        self._mark_as_completed()
        failure(_failure)

        if detail:
            quiet(detail)

        if message:
            error(message)

    def success(self, detail: Optional[str] = None) -> None:
        """Indicate that the task succeeded.

        Args:
            detail (str, optional): Optional detail to appear under the task status.
        """
        self._mark_as_completed()
        success(_success)

        if detail:
            quiet(detail)

    def skipped(self, detail: Optional[str] = None) -> None:
        """Indicate that the task was skipped.

        Args:
            detail (str, optional): Optional detail to appear under the task status. Defaults to None.
        """
        self._mark_as_completed()
        skipped(_skipped)

        if detail:
            quiet(detail)

    def _mark_as_completed(self) -> None:
        """Mark the task as completed.

        Raises:
            Exception: If the status of the task has already been indicated.
        """
        if self.completed:
            raise Exception('Status already completed!')

        self.completed = True


@contextmanager
def critical_task(label: str, max_width: int = 80) -> Generator[Status, None, None]:  # pragma: no cover
    with task(label, max_width, exc='print_and_exit') as status:
        yield status


def completed_task(
    label: str, result: Union[bool, Exception], max_width: int = 80, exc: str = _raise, detail: Optional[str] = None,
):
    with task(label, max_width=max_width, exc=exc) as status:
        if isinstance(result, Exception):
            raise result
        elif result:
            status.success(detail=detail)
        else:
            status.failure(detail=detail)


@contextmanager
def task(label: str, max_width: int = 80, exc: str = _raise) -> Generator[Status, None, None]:  # noqa: WPS231,WPS213
    """Print a task's name, followed by its outcome.

    The outcome of the task is provided by called the `success` or `failure`
    methods on the provided `status` object.

    Exceptions thrown in the yield block can be handled in mulitple ways via the `exc` argument:
    - 'raise', the exception is raised, nothing is printed (default)
    - 'print', the exception is printed, execution continues
    - 'exit', the process exits, nothing is printed
    - 'print_and_exit', the exception is printer, the process exits

    Args:
        label (str): The name of the task.
        max_width (int): The max width of the printed message.
        exc (str): The exception strategy.

    Yields:
        Status: The status object.

    Raises:
        Exception: Passes through the exception raised in the yield block.

    Examples:
        ```
        with console.status('My Task') as status:
            did_succeed = run_task()

            if did_succeed:
                status.success()
            else:
                status.failure()
        ```
    """
    assert exc in {_raise, _print, _exit, _print_and_exit}  # noqa: S101

    dim_dots = stylize('.' * (max_width - len(label)), colored.attr(_dim))
    bold_label = stylize(label, colored.attr(_bold))

    write(f'{bold_label}{dim_dots}', with_newline=False)

    work_status = Status()

    try:
        yield work_status
    except Exception as ex:
        if not work_status.completed:
            work_status.failure()

        if exc == _raise:
            raise ex
        elif exc == _print:
            error(_task_failed)
        elif exc == _exit:
            fail()
        else:
            fail(_task_failed)

    if not work_status.completed:
        work_status.success()


@contextmanager
def padding(**kwargs) -> Generator[None, None, None]:
    """Pad the output in newlines.

    Args:
        kwargs (Any): Options passed to `write`.

    Yields:
        None
    """
    newline(**kwargs)
    yield
    newline(**kwargs)


def success(message: str, **kwargs) -> None:
    """Write a bold green message.

    Args:
        message (str): The message.
        kwargs (Any): Options passed to `write`.
    """
    write(bold_green(message), **kwargs)


def failure(message: str, **kwargs) -> None:
    """Write a bold red message.

    Args:
        message (str): The message.
        kwargs (Any): Options passed to `write`.
    """
    write(bold_red(message), **kwargs)


def skipped(message: str, **kwargs) -> None:
    """Write a dimmed message.

    Args:
        message (str): The message.
        kwargs (Any): Options passed to `write`.
    """
    quiet(message, **kwargs)


def quiet(message: str, **kwargs) -> None:
    """Write a dimmed message.

    Args:
        message (str): The message.
        kwargs (Any): Options passed to `write`.
    """
    write(dim(message), **kwargs)


def loud(message: str, **kwargs) -> None:
    """Write a bold message.

    Args:
        message (str): The message.
        kwargs (Any): Options passed to `write`.
    """
    write(bold(message), **kwargs)


def header(message: str, **kwargs) -> None:
    """Write a bold message with padding.

    Args:
        message (str): The message.
        kwargs (Any): Options passed to `write`.
    """
    with padding(**kwargs):
        loud(message, **kwargs)


def newline(**kwargs) -> None:
    """Write a newline.

    Args:
        kwargs (Any): Options passed to `write`.
    """
    write('\n', with_newline=False, **kwargs)


def write(message: str, file: Optional[IO[str]] = None, with_newline: bool = True) -> None:
    """Write a message.

    Args:
        message (str): The message.
        file (IO): The file descriptor to which the message will be written.
        with_newline (bool): Whether to print a newline.
    """
    if not file:
        file = sys.stdout

    file.write(message)

    if with_newline:
        file.write('\n')

    file.flush()


def write_list(message: str, items: List, bullet: str = _bullet, **kwargs) -> None:
    """Prints a header and a list of items.

    Args:
        message (str): The header message.
        items (List): The list of items to print.
        bullet (str): The bullet character to use.
        kwargs (Any): Options passed to `write`.
    """
    with padding(**kwargs):
        header(message, **kwargs)

        for item in items:
            write(f' {bullet} {item}', **kwargs)


def fail(message: Optional[str] = None, **kwargs) -> None:
    """Prints an error message to stderr and exits the script with a non-zero exit code.

    Args:
        message (str): The message to print.
        kwargs (Any): Options passed to `write`.
    """
    if message:
        error(message, **kwargs)
    sys.exit(-1)


def error(message: str, **kwargs) -> None:
    """Prints an error message to stderr.

    Optionally print the stacktrace for the current exception.

    Args:
        message (str): The message to be displayed.
        kwargs (Any): Options passed to `write`.
    """
    error_prefix = bold_red(_error)
    message = bold(message)
    write(f'{error_prefix}: {message}', **{'file': sys.stderr, **kwargs})

    # If we're currently handling an exception, print the stacktrace
    _ex_type, ex, _trace = sys.exc_info()
    if ex:
        quiet(traceback.format_exc(), **{'file': sys.stderr, **kwargs})
